#ifndef __MONITOR_DT_H__
#define __MONITOR_DT_H__

#include "io.h"
#include "DE.h"

#if defined(FUNCPROTO) || __STDC__ || defined(__cplusplus) || defined(c_plusplus)
#ifndef ARGS
#define ARGS(args) args
#endif
#endif

#ifndef False
#define False 0
#endif

#ifndef True
#define True (!False)
#endif


#define UR_PROCESS_EXIT			209

typedef int (*EventHandlerProc)();

#define NUMBER_OF_EVENTS	18

/* default sensor numbers */
#define SIMPLE_COMMAND		0
#define FLOAT_COMMAND		1
#define NAMED_COMMAND		2
#define DM_THREAD_INIT          3 
#define DM_THREAD_FORK		4 
#define DM_THREAD_EXIT		5 
#define DM_THREAD_BEGIN_JOIN  	6 
#define DM_THREAD_END_JOIN	7 
#define DM_THREAD_DETACH	8 
#define DM_THREAD_SET_NAME	9 
#define DM_MUTEX_BEGIN_LOCK	10
#define DM_MUTEX_END_LOCK	11
#define DM_MUTEX_SET_NAME	12
#define DM_CONDITION_BEGIN_WAIT 13
#define DM_CONDITION_END_WAIT   14
#define DM_CONDITION_SET_NAME	15
#define DM_THREAD_OFF_RUNQ      16
#define DM_MONITOR_EXIT		17

/* special event */
#define SEVENT_PERTURBATION	99

typedef struct _URSimpleCommand {
    short       tag;
    short       type;
    short       proc;
    short       sensor;
    int     int_field;
    int     aux_int;
} URSimpleCommand, *URSimpleCommandPtr;

typedef struct _URFloatCommand {
    short       tag;
    short       type;
    short       proc;
    short       sensor;
    double  float_field;
    double  aux_float;
} URFloatCommand, *URFloatCommandPtr;

typedef struct _URNamedCommand {
    short       tag;
    short       type;
    short       proc;
    short       sensor;
    char    *file_name;
    char    *aux_name;
} URNamedCommand, *URNamedCommandPtr;


/* This is the header for each sensor information */
typedef struct _SimpleTraceDataRecord {
    int            type;
    int		   perturbation;
    unsigned long  timestamp;
    unsigned long  thread;
} SimpleTraceDataRecord, *SimpleTraceDataRecordPtr;

/**  special event PERTURBATION */
typedef struct _SEventPerturbationType {
    SimpleTraceDataRecord  header;
} SEventPerturbationType, *SEventPerturbationTypePtr;
    
/* default sensor data structures */

/**  sensor thread  init ***/
typedef struct _DM_Type_thread_init {
    SimpleTraceDataRecord  header;
    char           *name;
    int	           num_processors;
    int		   num_lms;
    int		   on_processor;
    unsigned long  timestamp_resolution;
} DM_Type_thread_init, *DM_Type_thread_initPtr;


/*** sensor thread_fork ***/

typedef struct _DM_Type_thread_fork {
    SimpleTraceDataRecord  header;
    char           *name;
    unsigned long  child_thread;
    int		   sequence_number;
    int		   on_processor;
} DM_Type_thread_fork, *DM_Type_thread_forkPtr;


/*** sensor thread_exit ***/

typedef struct _DM_Type_thread_exit {
    SimpleTraceDataRecord  header;
} DM_Type_thread_exit, *DM_Type_thread_exitPtr;


/*** sensor thread_begin_join ***/

typedef struct _DM_Type_thread_begin_join {
    SimpleTraceDataRecord  header;
    unsigned long  join_to;
} DM_Type_thread_begin_join, *DM_Type_thread_begin_joinPtr;


/*** sensor thread_end_join ***/

typedef struct _DM_Type_thread_end_join {
    SimpleTraceDataRecord  header;
    unsigned long  join_to;
} DM_Type_thread_end_join, *DM_Type_thread_end_joinPtr;


/*** sensor thread_detach ***/

typedef struct _DM_Type_thread_detach {
    SimpleTraceDataRecord  header;
    unsigned long  detach_from;
} DM_Type_thread_detach, *DM_Type_thread_detachPtr;


/*** sensor thread_set_name ***/

typedef struct _DM_Type_thread_set_name {
    SimpleTraceDataRecord  header;
    unsigned long the_thread;
    char  *name;
} DM_Type_thread_set_name, *DM_Type_thread_set_namePtr;


/*** sensor mutex_begin_lock ***/

typedef struct _DM_Type_mutex_begin_lock {
    SimpleTraceDataRecord  header;
    unsigned long  mutex;
    int            sequence_number;
} DM_Type_mutex_begin_lock, *DM_Type_mutex_begin_lockPtr;


/*** sensor mutex_end_lock ***/

typedef struct _DM_Type_mutex_end_lock {
    SimpleTraceDataRecord  header;
    unsigned long  mutex;
    int            sequence_number;
} DM_Type_mutex_end_lock, *DM_Type_mutex_end_lockPtr;


/*** sensor mutex_set_name ***/

typedef struct _DM_Type_mutex_set_name {
    SimpleTraceDataRecord  header;
    unsigned long  mutex;
    char           *name;
} DM_Type_mutex_set_name, *DM_Type_mutex_set_namePtr;


/*** sensor condition_begin_wait ***/

typedef struct _DM_Type_condition_begin_wait {
    SimpleTraceDataRecord  header;
    unsigned long  condition;
    int            sequence_number;
    unsigned long  mutex;
} DM_Type_condition_begin_wait, *DM_Type_condition_begin_waitPtr;


/*** sensor condition_end_wait ***/

typedef struct _DM_Type_condition_end_wait {
    SimpleTraceDataRecord  header;
    unsigned long  condition;
    int            sequence_number;
    unsigned long  mutex;
} DM_Type_condition_end_wait, *DM_Type_condition_end_waitPtr;


/*** sensor condition_set_name ***/

typedef struct _DM_Type_condition_set_name {
    SimpleTraceDataRecord  header;
    unsigned long  condition;
    char           *name;
} DM_Type_condition_set_name, *DM_Type_condition_set_namePtr;


typedef struct _DM_Readyq {
    SimpleTraceDataRecord  header;
} DM_Type_readyq, *DM_Type_readyqPtr;


/*** monitor thread exit ***/

typedef struct _DM_Type_monitor_exit {
    SimpleTraceDataRecord  header;
    int                    lm_number;
} DM_Type_monitor_exit, *DM_Type_monitor_exitPtr;


typedef struct _DM_Type_thread_off_runq {
    SimpleTraceDataRecord  header;
    unsigned long thread_off_runq;
} DM_Type_thread_off_runq, *DM_Type_thread_off_runqPtr;

/*
typedef union _Event_Data_Type {
    SimpleTraceDataRecord       header;
    SEventPerturbationType      perturbation_event;
    DM_Type_thread_init    	thread_init;
    DM_Type_thread_fork    	thread_fork;
    DM_Type_thread_exit    	thread_exit;
    DM_Type_thread_begin_join 	thread_join;
    DM_Type_thread_detach  	thread_detach;
    DM_Type_thread_set_name 	thread_set_name;
    DM_Type_mutex_begin_lock 	mutex_lock;

    DM_Type_mutex_set_name 	mutex_set_name;
    DM_Type_condition_begin_wait condition_wait;
    DM_Type_condition_set_name 	condition_set_name;
    DM_Type_thread_off_runq 	thread_off_runq;
    DM_Type_monitor_exit    	monitor_exit;
} Event_Data_Type, *Event_Data_TypePtr; */

/* external functions */
extern int CMonInterfInit ARGS((int*, char**));
extern int ReadAndProcessInput ();
extern EventHandlerProc RegisterEventHandler ARGS((int, EventHandlerProc));
    

#endif
