/***************************************************************/
/*							       */
/*			     POLKA_GC.H			       */
/*							       */
/***************************************************************/
/*       Copyright 1993 Georgia Institute of Technology 
                     -- John T. Stasko                         */

/**************************************************************/
/*****************	LOCAL data types     ******************/
/**************************************************************/

typedef struct {
   char *name;
   XFontStruct *xfs;
   } GC_FONT;

typedef struct {
   int  width;		/* 0, 2, 4 		      */
   int  line_style;	/* LineSolid or LineOnOffDash */
   char dash_list[2];
   } GC_LINE;



/**************************************************************/
/*****************	GLOBAL variables     ******************/
/**************************************************************/

/**************************************************************/
/*****************	LOCAL variables      ******************/
/**************************************************************/

#ifdef ultrix
#define DEFAULTFONT      "fixed"
#else
#define DEFAULTFONT	 "8x13"
#endif
#define MAXCOLORS	 256
#define NUMDEFAULTCOLORS  8
#define MAXFONTS         20
#define MAXFILLS	 41
#define NUMDEFAULTFILLS  41
#define MAXLINESTYLES     9
#define UNINITIALIZED	 -1

static Pixel  _colors[MAXCOLORS];
static const char  *_defaultName[NUMDEFAULTCOLORS] = {
		"white", "yellow", "green",  "blue",
		"orange", "red",   "maroon", "black" };
static char  *_colorName[MAXCOLORS];

static GC_FONT _font[MAXFONTS] = { 0 };
static GC_LINE _lineStyles[MAXLINESTYLES] = {
		  {0,LineSolid,    {4,4}}, {0,LineOnOffDash,{8,8}},
		  {0,LineOnOffDash,{2,2}}, {2,LineSolid,    {4,4}},
		  {2,LineOnOffDash,{8,8}}, {2,LineOnOffDash,{2,2}},
		  {4,LineSolid,    {4,4}}, {4,LineOnOffDash,{8,8}},
		  {4,LineOnOffDash,{2,2}} };

static Pixmap _fillPixmap[MAXFILLS];

static GC  _gc;
static int _curFill, _curLS, _curColor, _curFont;
static int _numColorsAvail, _numFillsAvail, _numFontsAvail;

static char _fillStyles[][8] = {
   {(char)0x00, (char)0x00, (char)0x00, (char)0x00, 
            (char)0x00, (char)0x00, (char)0x00, (char)0x00},
   {(char)0x01, (char)0x00, (char)0x00, (char)0x00, 
            (char)0x10, (char)0x00, (char)0x00, (char)0x00},
   {(char)0x01, (char)0x00, (char)0x00, (char)0x00, 
            (char)0x11, (char)0x00, (char)0x00, (char)0x00},
   {(char)0x11, (char)0x00, (char)0x04, (char)0x00, 
            (char)0x11, (char)0x00, (char)0x00, (char)0x00},
   {(char)0x11, (char)0x00, (char)0x04, (char)0x00, 
            (char)0x11, (char)0x00, (char)0x40, (char)0x00},
   {(char)0x11, (char)0x00, (char)0x44, (char)0x00, 
            (char)0x11, (char)0x00, (char)0x44, (char)0x00},
   {(char)0x11, (char)0x00, (char)0x45, (char)0x00, 
            (char)0x11, (char)0x00, (char)0x54, (char)0x00},
   {(char)0x11, (char)0x00, (char)0x45, (char)0x00, 
            (char)0x11, (char)0x00, (char)0x55, (char)0x00},
   {(char)0x15, (char)0x00, (char)0x55, (char)0x00, 
            (char)0x11, (char)0x00, (char)0x55, (char)0x00},
   {(char)0x15, (char)0x00, (char)0x55, (char)0x00, 
            (char)0x51, (char)0x00, (char)0x55, (char)0x00},
   {(char)0x55, (char)0x00, (char)0x55, (char)0x00, 
            (char)0x55, (char)0x00, (char)0x55, (char)0x00},
   {(char)0x55, (char)0x02, (char)0x55, (char)0x00, 
            (char)0x55, (char)0x20, (char)0x55, (char)0x00},
   {(char)0x55, (char)0x02, (char)0x55, (char)0x00, 
            (char)0x55, (char)0x22, (char)0x55, (char)0x00},
   {(char)0x55, (char)0x22, (char)0x55, (char)0x08, 
            (char)0x55, (char)0x22, (char)0x55, (char)0x00},
   {(char)0x55, (char)0x22, (char)0x55, (char)0x08, 
            (char)0x55, (char)0x22, (char)0x55, (char)0x80},
   {(char)0x55, (char)0x22, (char)0x55, (char)0x88, 
            (char)0x55, (char)0x22, (char)0x55, (char)0x88},
   {(char)0x55, (char)0x22, (char)0x55, (char)0x8a, 
            (char)0x55, (char)0x22, (char)0x55, (char)0xa8},
   {(char)0x55, (char)0x22, (char)0x55, (char)0x8a, 
            (char)0x55, (char)0x22, (char)0x55, (char)0xaa},
   {(char)0x55, (char)0x2a, (char)0x55, (char)0xaa, 
            (char)0x55, (char)0x22, (char)0x55, (char)0xaa},
   {(char)0x55, (char)0x2a, (char)0x55, (char)0xaa, 
            (char)0x55, (char)0xa2, (char)0x55, (char)0xaa},
   {(char)0x55, (char)0xaa, (char)0x55, (char)0xaa, 
            (char)0x55, (char)0xaa, (char)0x55, (char)0xaa},
   {(char)0x55, (char)0xab, (char)0x55, (char)0xaa, 
            (char)0x55, (char)0xba, (char)0x55, (char)0xaa},
   {(char)0x55, (char)0xab, (char)0x55, (char)0xaa, 
            (char)0x55, (char)0xbb, (char)0x55, (char)0xaa},
   {(char)0x55, (char)0xbb, (char)0x55, (char)0xae, 
            (char)0x55, (char)0xbb, (char)0x55, (char)0xaa},
   {(char)0x55, (char)0xbb, (char)0x55, (char)0xae, 
            (char)0x55, (char)0xbb, (char)0x55, (char)0xea},
   {(char)0x55, (char)0xbb, (char)0x55, (char)0xee, 
            (char)0x55, (char)0xbb, (char)0x55, (char)0xee},
   {(char)0x55, (char)0xbb, (char)0x55, (char)0xef, 
            (char)0x55, (char)0xbb, (char)0x55, (char)0xee},
   {(char)0x55, (char)0xbb, (char)0x55, (char)0xef, 
            (char)0x55, (char)0xbb, (char)0x55, (char)0xff},
   {(char)0x55, (char)0xbf, (char)0x55, (char)0xff, 
            (char)0x55, (char)0xbb, (char)0x55, (char)0xff},
   {(char)0x55, (char)0xbf, (char)0x55, (char)0xff, 
            (char)0x55, (char)0xfb, (char)0x55, (char)0xff},
   {(char)0x55, (char)0xff, (char)0x55, (char)0xff, 
            (char)0x55, (char)0xff, (char)0x55, (char)0xff},
   {(char)0x57, (char)0xff, (char)0x55, (char)0xff, 
            (char)0x75, (char)0xff, (char)0x55, (char)0xff},
   {(char)0x57, (char)0xff, (char)0x55, (char)0xff, 
            (char)0x77, (char)0xff, (char)0x55, (char)0xff},
   {(char)0x77, (char)0xff, (char)0x5d, (char)0xff, 
            (char)0x77, (char)0xff, (char)0x55, (char)0xff},
   {(char)0x77, (char)0xff, (char)0x5d, (char)0xff, 
            (char)0x77, (char)0xff, (char)0xd5, (char)0xff},
   {(char)0x77, (char)0xff, (char)0xdd, (char)0xff, 
            (char)0x77, (char)0xff, (char)0xdd, (char)0xff},
   {(char)0x77, (char)0xff, (char)0xdf, (char)0xff, 
            (char)0x77, (char)0xff, (char)0xfd, (char)0xff},
   {(char)0x77, (char)0xff, (char)0xdf, (char)0xff, 
            (char)0x77, (char)0xff, (char)0xff, (char)0xff},
   {(char)0x7f, (char)0xff, (char)0xff, (char)0xff, 
            (char)0x77, (char)0xff, (char)0xff, (char)0xff},
   {(char)0x7f, (char)0xff, (char)0xff, (char)0xff, 
            (char)0xf7, (char)0xff, (char)0xff, (char)0xff},
   {(char)0xff, (char)0xff, (char)0xff, (char)0xff,
            (char)0xff, (char)0xff, (char)0xff, (char)0xff}
};
