/***************************************************************/
/*							       */
/*			     POLKAGC.C			       */
/*							       */
/***************************************************************/
/*       Copyright 1993 Georgia Institute of Technology 
                     -- John T. Stasko                         */

#include "polka_local.H"
#include "polka_gc.H"


static int already_loaded(COLOR);

GC
inq_gc()
{
  return _gc;
}


int 
inq_font() 
{
   return _curFont;
}


int 
inq_color() 
{
   return _curColor;
}




void
SetBackground(COLORINDEX c)
{
   XSetBackground(_display, _gc, GetXColor(c));
}



COLORINDEX
load_color(COLOR name)
   {
   XColor color;
   int in_it;

   if (!name) return _curColor;
   in_it = already_loaded(name);
   if (in_it != -1) return in_it;

   if (_color_screen) {
      if (XParseColor(_display, DefaultColormap(_display,
			      DefaultScreen(_display)), name,
			      &color)) {
         XAllocColor(_display, DefaultColormap(_display,
                                 DefaultScreen(_display)),
                         &color);
	 _colors[_numColorsAvail] = color.pixel;
	 return _numColorsAvail++;
	 }
      else {
	 POLKA_complain("load_color: color not available", name);
         free(_colorName[_numColorsAvail]);
	 return BLACK;
	 }
      }
   else {
      return _numColorsAvail++;
      }
   }


Pixel
GetXColor(COLORINDEX color)
   {
   if (color < 0 || color >= _numColorsAvail) {
      POLKA_complain("GetXColor", "undefined color");
      return _colors[WHITE];
      }

   if (_color_screen) {
      return( _colors[color] );
      }
   else {
      return( (color == WHITE
                  ? WhitePixel(_display, DefaultScreen(_display))
                  : BlackPixel(_display, DefaultScreen(_display))));
	 }
}



int
already_loaded(COLOR c)
{
   int i;

   for (i=0; (i<=_numColorsAvail-1) && (i<MAXCOLORS); ++i)
      if (streql((char*)c,_colorName[i]))
         return(i);
   if (i==MAXCOLORS) {
      POLKA_complain("already_loaded","Max number of colors used.  Loading black");
      return (BLACK);
   }
   _colorName[_numColorsAvail] = (char *) malloc(strlen(c) + 1);
   strcpy(_colorName[_numColorsAvail],c);
   return(-1);
}


COLORINDEX
set_color(COLORINDEX color)
   {
   int old = _curColor;
   Pixel fg;

   if (color == _curColor) return _curColor;
   if (color < 0 || color >= _numColorsAvail) {
      POLKA_complain("set_color", "undefined color");
      return _curColor;
      }

   if (_color_screen) {
      fg = _colors[color];
      }
   else {
      fg = (color == WHITE
                  ? WhitePixel(_display, DefaultScreen(_display))
                  : BlackPixel(_display, DefaultScreen(_display)));
      if (color == WHITE || color == BLACK)
         XSetFillStyle(_display, _gc, FillSolid);
      else  /* 8 distinct fills */
         fill_style((int)(color*5.7 + 0.5));
      }

      XSetForeground(_display, _gc, fg);
      return (old != UNINITIALIZED ? old : _curColor);
   }



LINE_STYLE 
line_style(LINE_STYLE style)
   {
   int old = _curLS;
   
   if (style < 0 || style > MAXLINESTYLES) {
      POLKA_complain("line_style", "undefined line style");
      return _curLS;
      }
   if (style == _curLS) return _curLS;
   _curLS = style;

   if (style == SOLID)
      XSetLineAttributes(_display, _gc, 0, LineSolid,
			 CapButt, JoinMiter);
   else {
      XSetLineAttributes(_display, _gc, _lineStyles[style].width,
   		         _lineStyles[style].line_style, CapButt, JoinMiter);
      XSetDashes(_display, _gc, 0, _lineStyles[style].dash_list, 2);
      }
   return old;
   }



FILL_STYLE 
fill_style(FILL_STYLE fill)
   {
   int old = _curFill;

   if (fill < 0 || fill >= _numFillsAvail) {
      POLKA_complain("fill_style", "undefined fill style\n\n");
      return _curFill;
      }
   if (fill == _curFill) return _curFill;
   _curFill = fill;

   /* For mono screens, colors are represented by fillstyles  */
   /*    so any change to fillstyle must alert the color      */
   /*    routine that the current "color" is NOT what it used */
   /*    to be.  Force TANGO_color() to reload the color.     */

   if (!_color_screen) _curColor = UNINITIALIZED;

   /* Update the GC with the new fill style */

   if (_fillPixmap[fill] != None) {
      XSetFillStyle(_display, _gc, FillOpaqueStippled);
      XSetStipple(_display, _gc, _fillPixmap[fill]);
      }
   else /* Problem getting particular fill Pixmap */
      XSetFillStyle(_display, _gc, FillSolid);
   
   return old;
   }


int 
load_font(char *fontname)
   {
   XFontStruct *xfs;
   int newfont = 0;
   int i,found;

   if (!fontname) return newfont;
   if (_numFontsAvail >= MAXFONTS) {
      POLKA_complain("load_font","max number of fonts already used");
      return newfont;
      }
   
   /* Make sure we don't reload an already-used font */

   for (found = i = 0; !found && i < _numFontsAvail; i += !found)
      found = !strcmp(_font[i].name, fontname);
   if (found) return i;

   /* Font hasn't been loaded yet, load it */

   xfs = XLoadQueryFont(_display, fontname);
   if (xfs == None)	/* Problem loading font, use default */
      POLKA_complain("load_font","font not available");
   else {
      _font[_numFontsAvail].name = (char *) malloc(strlen(fontname) + 1);
      strcpy(_font[_numFontsAvail].name, fontname);
      _font[_numFontsAvail].xfs = xfs;
      newfont = _numFontsAvail++;
      }

   return newfont;
}


void 
text_info(Font fontid,char *str,int *xext,int *yext,int *xofs,int *yofs)
   {
   XCharStruct over;
   int dir, asc, des;

   if (fontid < 0 || fontid >= _numFontsAvail) {
      POLKA_complain("text_info","undefined font");
      *xext = *yext = *xofs = *yofs = 0;
      }
   else if (str) {
      XTextExtents(_font[fontid].xfs, str, strlen(str), &dir, &asc, &des, &over);
      *xext = over.rbearing;
      *yext = over.ascent + over.descent + 1;
      *xofs = -over.lbearing;
      *yofs = (over.descent > des ? over.descent : des);
      }
   else
      *xext = *yext = *xofs = *yofs = 0;
   }


int 
set_font(int fontid)
   {
   int old = _curFont;

   if (fontid == _curFont) return _curFont;
   if (fontid < 0 || fontid >= _numFontsAvail) {
      POLKA_complain("set_font","undefined font");
      return _curFont;
      }

   _curFont = fontid;
   XSetFont(_display, _gc, _font[fontid].xfs->fid);

   return old;
   }



void 
initGC() {
   int    i;
   COLOR_MAPPER *gcc;

   /* Init fill styles ... create pixmaps */
   for (i = 0; i < NUMDEFAULTFILLS; i++) {
      _fillPixmap[i] = XCreateBitmapFromData(_display,
				DefaultRootWindow(_display),
				_fillStyles[i], 8, 8);
      if (_fillPixmap[i] == None) {
/*         POLKA_complain("POLKA_initGC", "unable to create fillstyle");  */
         /* recover somehow...here */
         }
      }
   for ( ; i < MAXFILLS; i++)
      _fillPixmap[i] = None;
   _numFillsAvail = NUMDEFAULTFILLS;

   /* Init colors */
   for (i = 0; i < NUMDEFAULTCOLORS; i++)
      load_color((char*)_defaultName[i]);
   for ( ; i < MAXCOLORS; i++)
      _colors[i] = UNINITIALIZED;
    _numColorsAvail = NUMDEFAULTCOLORS;

   /* Init fonts */
   for (i = 0; i < MAXFONTS; i++) {
      _font[i].name = NULL;
      _font[i].xfs = NULL;
      }
   _numFontsAvail = 0;

   /* Create GC */
   _gc = XCreateGC(_display, DefaultRootWindow(_display), 0, 0);
   if (_gc == None) {
      POLKA_complain("initGC", "unable to create GC");
      /* recover somehow...here */
      exit(1);
      }


   /* Init GC - white background, default font, black */
   /*		foreground, solid fill, solid line    */

   XSetBackground(_display, _gc, WhitePixel(_display,DefaultScreen(_display)));

   _color_list = new(COLOR_MAPPER);  // used for changing objects
   _color_list->pathval = 0.0;   // to different colors
   _color_list->colornum = BLACK;
   strcpy(_color_list->name,"black");
   gcc = new(COLOR_MAPPER);
   gcc->next = NULL;
   gcc->pathval = 0.001;  
   gcc->colornum = WHITE;
   strcpy(gcc->name,"white");
   _color_list->next = gcc;

   _curFont = _curColor = _curFill = _curLS = UNINITIALIZED;
   set_font(load_font((char*)DEFAULTFONT));
   fill_style(40);
   set_color(BLACK);
   line_style(SOLID);
}
