/* Required definitions. */
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

/* Various required header files. */
#include <linux/config.h> 
#include <linux/module.h> 
#include <linux/version.h> 
#include <linux/kernel.h> 
#include <linux/init.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/socket.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <asm/semaphore.h>
#include "kernel/kpbio.h"

/* Socket send function. */
int my_sock_sendmsg(struct socket *sock, char *buf, int len) 
{
  struct msghdr msg;
  struct iovec iov;

  if (len == 0)
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  msg.msg_flags = 0;
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  iov.iov_base = (void *)buf;
  iov.iov_len = len;
  if (sock->type == SOCK_SEQPACKET)
    msg.msg_flags |= MSG_EOR;

/*
#ifndef KECHO
  if (rt_sock_sendmsg(sock, &msg, len, jiffies, 0) <= 0) {
    schedule_timeout(1);
    rt_sock_sendmsg(sock, &msg, len, jiffies, 0);
  }
#else
*/
  if (sock_sendmsg(sock, &msg, len) <= 0) {
    schedule_timeout(1);
    sock_sendmsg(sock, &msg, len);
  }
/*
#endif
*/
  return len;
}

/* Socket send function. */
int my_sock_sendmsgv(struct socket *sock, struct iovec *iov, int len) 
{
  struct msghdr msg;
  int tot_len = 0;
  int i;

  if (len == 0)
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  msg.msg_flags = 0;
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = iov;
  msg.msg_iovlen = len;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  if (sock->type == SOCK_SEQPACKET)
    msg.msg_flags |= MSG_EOR;

  for (i = 0; i < len; i++)
    tot_len += iov[i].iov_len;

/*
#ifndef KECHO
  if (rt_sock_sendmsg(sock, &msg, tot_len, jiffies, 0) <= 0) {
    schedule_timeout(1);
    rt_sock_sendmsg(sock, &msg, tot_len, jiffies, 0);
  }
#else
*/
  if (sock_sendmsg(sock, &msg, tot_len) <= 0) {
    schedule_timeout(1);
    sock_sendmsg(sock, &msg, tot_len);
  }
/*
#endif
*/
  return tot_len;
}

/* Socket read function. */
int my_sock_recvmsg(struct socket *sock, char *buf, int len) 
{
  struct msghdr msg;
  struct iovec iov;

  if (len == 0)
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  msg.msg_flags = 0;
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  iov.iov_base = (void *)buf;
  iov.iov_len = len;

  return sock_recvmsg(sock, &msg, len, 0);
}

/* Socket read function. */
int my_sock_recvmsgv(struct socket *sock, struct iovec *iov, int len) 
{
  struct msghdr msg;
  size_t tot_len = 0;
  int i;

  if (len == 0)
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  for (i = 0; i < len; i++)
    tot_len += iov[i].iov_len;

  msg.msg_flags = 0;
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = iov;
  msg.msg_iovlen = len;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;

  return sock_recvmsg(sock, &msg, tot_len, 0);
}

int sys_gethostname(char *name, int len)
{
  int i, errno;

  if (len < 0)
    return -EINVAL;
  down_read(&uts_sem);
  i = 1 + strlen(system_utsname.nodename);
  if (i > len)
    i = len;
  errno = 0;
  memcpy(name, system_utsname.nodename, i);
  up_read(&uts_sem);
  return errno;
}

/* arch/i386/kernel/entry.S */
extern void *sys_call_table[];

/* List of system calls used in the module. */
int     (*sys_exit)       (int error_code);
int     (*sys_fork)       (struct pt_regs regs);
__kernel_ssize_t (*sys_kread) (unsigned int fd, char *buf, size_t count);
__kernel_ssize_t (*sys_kwrite)(unsigned int fd, const char *buf, size_t count);
int     (*sys_kopen)      (const char *filename, int flags, int mode);
int     (*sys_kclose)     (unsigned int fd);
int     (*sys_waitpid)    (__kernel_pid_t pid, unsigned int *stat_addr, 
                           int options);
int     (*sys_unlink)     (const char *pathname);
int     (*sys_time)       (int *tloc);
int     (*sys_ni_syscall) (void);
int     (*sys_getpid)     (void);
int     (*sys_getuid)     (void);
unsigned int (*sys_alarm) (unsigned int seconds);
int     (*sys_kill)       (int pid, int sig);
int     (*sys_pipe)       (unsigned long *fildes);
unsigned long (*sys_signal) (int sig, __sighandler_t handler);
int     (*sys_sigaction)  (int sig, const struct old_sigaction *act,
                           struct old_sigaction *oact);
int     (*sys_setrlimit)  (unsigned int resource, struct rlimit *rlim);
int     (*sys_select)     (int n, fd_set *inp, fd_set *outp,
                          fd_set *exp, struct timeval *tvp);
int     (*sys_truncate)   (const char *path, unsigned long length);
__kernel_ssize_t (*sys_readv) (unsigned long fd, const struct iovec *vector,
                           unsigned long count);
__kernel_ssize_t (*sys_writev)  (unsigned long fd, const struct iovec *vector,
                           unsigned long count);
int     (*sys_rt_sigaction) (int sig, const struct sigaction *act, struct
                             sigaction *oact, size_t sigsetsize);

int     (*sys_socket)      (int family, int type, int protocol);
int     (*sys_connect)     (int fd, struct sockaddr *uservaddr, int addrlen);
long    (*sys_kfcntl)       (unsigned int fd, unsigned int cmd, 
			    unsigned long arg);


/* strdup function */
char *strdup(char *s)
{
  char *p;

  p = (char *)kmalloc(strlen(s)+1, GFP_KERNEL);
  if (p != NULL)
    strcpy(p, s);
  return p;
}

int atoi(const char *nptr)
{
  return (int) strtol(nptr, (char **) NULL, 10);
}

/* Initialize module. */
static int __init kpbio_init(void) 
{
  /* Get system table entries. */
  sys_exit         =   sys_call_table[__NR_exit];
  sys_fork         =   sys_call_table[__NR_fork];
  sys_kread         =   sys_call_table[__NR_read];
  sys_kwrite        =   sys_call_table[__NR_write];
  sys_kopen        =   sys_call_table[__NR_open];
  sys_kclose       =   sys_call_table[__NR_close];
  sys_waitpid      =   sys_call_table[__NR_waitpid];
  sys_unlink       =   sys_call_table[__NR_unlink];
  sys_time         =   sys_call_table[__NR_time];
  sys_ni_syscall   =   sys_call_table[__NR_break];
  sys_getpid       =   sys_call_table[__NR_getpid];
  sys_getuid       =   sys_call_table[__NR_getuid];
  sys_alarm        =   sys_call_table[__NR_alarm];
  sys_kill         =   sys_call_table[__NR_kill];
  sys_pipe         =   sys_call_table[__NR_pipe];
  sys_signal       =   sys_call_table[__NR_signal];
  sys_sigaction    =   sys_call_table[__NR_sigaction];
  sys_setrlimit    =   sys_call_table[__NR_setrlimit];
  sys_select       =   sys_call_table[__NR_select];
  sys_truncate     =   sys_call_table[__NR_truncate];
  sys_readv        =   sys_call_table[__NR_readv];
  sys_writev       =   sys_call_table[__NR_writev];
  sys_rt_sigaction =   sys_call_table[__NR_rt_sigaction];
  sys_kfcntl        =   sys_call_table[__NR_fcntl];

  return 0;
}

/* Remove module. */
static void __exit kpbio_cleanup(void) 
{
}

module_init(kpbio_init);
module_exit(kpbio_cleanup); 

