
#include <stdio.h>
#include "config.h"
#include "sensor.h"

char*
TypeString(type)
    int type;
{
    switch (type) {
    case CHAR_TYPE:
	return ("char           ");
    case UCHAR_TYPE:
	return ("unsigned char  ");
    case SHORT_TYPE:
	return ("short          ");
    case USHORT_TYPE:
	return ("unsigned short ");
    case INT_TYPE:
	return ("int            ");
    case UINT_TYPE:
	return ("unsigned int   ");
    case LONG_TYPE:
	return ("long           ");
    case ULONG_TYPE:
	return ("unsigned long  ");
    case FLOAT_TYPE:
	return ("float          ");
    case DOUBLE_TYPE:
	return ("double         ");
    case STRING_TYPE:
	return ("char*          ");
    }
    return NULL;
}

void
IOTypeString(type, type_name_str, type_name_iostr)
    int type;
    char **type_name_str, **type_name_iostr;
{
    *type_name_str = *type_name_iostr = NULL;
    switch (type) {
    case CHAR_TYPE:
	*type_name_str = *type_name_iostr = "char";
	break;
    case UCHAR_TYPE:
	*type_name_str = "unsigned char";
	*type_name_iostr = "unsigned char";
	break;
    case SHORT_TYPE:
	*type_name_str = "short";
	*type_name_iostr = "integer";
	break;
    case USHORT_TYPE:
	*type_name_str = "unsigned short";
	*type_name_iostr = "unsigned integer";
	break;
    case INT_TYPE:
	*type_name_str = "int";
	*type_name_iostr = "integer";
	break;
    case UINT_TYPE:
	*type_name_str = "unsigned int";
	*type_name_iostr = "unsigned integer";
	break;
    case LONG_TYPE:
	*type_name_str = "long";
	*type_name_iostr = "integer";
	break;
    case ULONG_TYPE:
	*type_name_str = "unsigned long";
	*type_name_iostr = "unsigned integer";
	break;
    case FLOAT_TYPE:
	*type_name_str = "float";
	*type_name_iostr = "float";
	break;
    case DOUBLE_TYPE:
	*type_name_str = "double";
	*type_name_iostr = "double";
	break;
    case STRING_TYPE:
	*type_name_str = "char*";
	*type_name_iostr = "string";
	break;
    }
}

LocationP
LocationInit(loc_type, loc_name, loc_lines)
    LocationType  loc_type;
    char          *loc_name;
    NumberListP   loc_lines;
{
    NumberListP  head;
    int  i, num_lines = 0;
    LocationP loc_p = (LocationP) malloc(sizeof(Location));
    loc_p->type = loc_type;
    if (loc_name) loc_p->name = strdup(loc_name);
    else loc_p->name = NULL;
    head = loc_lines;
    while (head) { num_lines++; head = head->next; }
    if (num_lines > 0) {
        loc_p->lines = (int*) malloc(sizeof(int)*(num_lines+1));
        loc_p->lines[0] = num_lines;
        head = loc_lines;
        for (i = 1; i <= num_lines; i++) {
	    loc_p->lines[i] = head->number;
	    head = head->next;
        }
    }
    loc_p->prev = loc_p->next = NULL;
    return loc_p;
}


AttributeP
AttributeInit(attr_type, attr_array_size, attr_name, attr_expr)
    AttributeType  attr_type;
    int		   attr_array_size;
    char	   *attr_name, *attr_expr;
{
    AttributeP attr_p = (AttributeP) malloc(sizeof(Attribute));
    attr_p->type = attr_type;
    attr_p->arraySize = attr_array_size;
    if (attr_name) attr_p->name = strdup(attr_name);
    else attr_p->name = NULL;
    if (attr_expr) attr_p->expression = strdup(attr_expr);
    else attr_p->expression = NULL;
    attr_p->prev = attr_p->next = NULL;
    return attr_p;
}

SensorP
SensorInit(sensor_type, sensor_name, attrs, samp_attrs, locs, pols)
    int  sensor_type;
    char *sensor_name;
    AttributeP	attrs, samp_attrs;
    LocationP   locs;
    PolicyP     pols;
{
    AttributeP attr_p;
    LocationP  loc_p;
    PolicyP    pol_p;
    SensorP    sensor_p = (SensorP) malloc(sizeof(Sensor));
    sensor_p->sensorType = sensor_type;
    if (sensor_name) sensor_p->name = strdup(sensor_name);
    else sensor_p->name = NULL;
    sensor_p->numAttributes = 0;
    attr_p = sensor_p->attributes = attrs;
    while (attr_p) {
	sensor_p->numAttributes++;
	attr_p = attr_p->next;
    }
    sensor_p->numSampAttributes = 0;
    if (sensor_type == SAMPLE_SENSOR) {
	attr_p = sensor_p->samp_attrs = samp_attrs;
	while (attr_p) {
	    sensor_p->numSampAttributes++;
	    attr_p = attr_p->next;
	}
    }
    sensor_p->numLocations = 0;
    loc_p = sensor_p->locations = locs;
    while (loc_p) {
	sensor_p->numLocations++;
	loc_p = loc_p->next;
    }
    sensor_p->numPolicies = 0;
    pol_p = sensor_p->policies = pols;
    while (pol_p) {
	sensor_p->numPolicies++;
	pol_p = pol_p->next;
    }
    sensor_p->prev = sensor_p->next = NULL;
    sensor_p->have_array_fld = False;
    return sensor_p;
}
